#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <sys/select.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <ctype.h>
#include <string.h>

#include "POD.h"

int POD::SetDeviceFile(int DeviceNum)
{
   int StringLength;

   DeviceNumber = DeviceNum;
#ifndef CYGWIN
   StringLength = (int)(log10((DeviceNum + 1.0))) + 13;
   DeviceFile = (char *)(calloc(StringLength, sizeof(char)));
   sprintf(DeviceFile, "/dev/ttyUSB%d", DeviceNum);
#else
   StringLength = (int)(log10((DeviceNum + 5.0 + 1.0))) + 10;
   DeviceFile = (char *)(calloc(StringLength, sizeof(char)));
   sprintf(DeviceFile, "/dev/com%d", DeviceNum + 5);
#endif
   return(0);
}

int POD::Ping()
{
   int Result;
   unsigned char Data[10];
   int Response;

   Result = SendCommand(POD_COMMAND_PING);
   Result = ReceiveBytes(1, Data);
   Response = Data[0];
   if (Response == POD_RESPONSE_PONG) return(1);
   return(0);
}

int POD::DetectPODType()
{
   int Result;
   unsigned char Data[10];
   int Response;

   SendCommand(POD_COMMAND_GETTYPE);
   Result = ReceiveBytes(3, Data);
   PODType = (Data[1] << 8 ) | Data[2];
   Response = Data[0];
#if (DEBUG > 2)
   printf("Got a PODType of 0x%04d and Response of 0x%02d\n", PODType, Response);
#endif
   if (Response == POD_RESPONSE_ACK) return(PODType);
   PODType = POD_TYPE_UNKNOWN;
   return(0);
}

int POD::SendByte(unsigned char Data)
{
   unsigned char NewData[2];
   NewData[0] = Data;
   return(SendBytes(1, NewData));
}

int POD::SendBytes(int Length, int Data)
{
   SendByte((Data >> 8) & 0xFF);
   SendByte(Data & 0xFF);
   return(0);
}

int POD::SendCommand(unsigned int Command)
{
   unsigned char Data[2];
   Data[0] = (Command >> 8) & 0xFF;
   Data[1] = Command & 0xFF;
   return(SendBytes(2, Data));
}

int POD::GetInfo()
{
   int Result;
   unsigned char Data[200];
   int Response;
   int URL_Length;

   Result = SendCommand(POD_COMMAND_GETINFO);
   Result = ReceiveBytes(1, Data);
   Response = Data[0];
   if (Response != POD_RESPONSE_ACK)
   {
      printf("Hrm.  Bad response\n");
   }
   Result = ReceiveBytes(2, Data); /* this should be the POD Type */

   Version = (char *)(calloc(24, sizeof(char)));
   Result = ReceiveBytes(24, Data); /* read the version info */
   strncpy(Version, (char *)Data, 24);

   Result = ReceiveBytes(1, Data);
   URL_Length = Data[0];

   URL = (char *)(calloc(URL_Length, sizeof(char)));
   Result = ReceiveBytes(URL_Length, Data);
   strncpy(URL, (char *)Data, URL_Length);
   return(0);
}

POD::POD(int DeviceNum, int PortNum) /* called when PODEnv has already opened the device */
{
#if (DEBUG_OBJECTS)
   printf("In POD (%d, %d)\n", DeviceNum, PortNum);
#endif
   Port = PortNum;
   SetDeviceFile(DeviceNum);
   if (Ping()) Valid = 1;
   DetectPODType();
   GetInfo();
}

POD::POD(int DeviceNum) /* this constructor will open DeviceNum */
{
   int Result;
#if (DEBUG_OBJECTS)
   printf("In POD (%d)\n", DeviceNum);
#endif
   SetDeviceFile(DeviceNum);
   /* open the port and configure it */
   Result = OpenDevice();
   if (Result == POD_ERROR_BADDEVICE)
   {
      PODType = POD_TYPE_NONE;
      URL = NULL;
      Version = NULL;
      DeviceNumber = -1;
      Valid = 0;
      return;
   }
   SetTerminalSettings();
//   usleep(100000); /* apparently, the device needs some time to itself, otherwise it won't communicate right */
   usleep(10000); /* apparently, the device needs some time to itself, otherwise it won't communicate right */
   if (Ping()) Valid = 1;
   DetectPODType();
   GetInfo();
}

POD::POD()
{
   /* can't do much, so lets just open device 0 */
#if (DEBUG_OBJECTS)
   printf("In POD ()\n");
#endif
   POD(0);
}

int POD::GetPort()
{
   return(Port);
}

int POD::GetDeviceNumber()
{
   return(DeviceNumber);
}

int POD::GetType()
{
   return(PODType);
}

int POD::GetType(char *PODTypeString)
{
   int Counter;
   typedef struct
   {
      int TypeNum;
      char *String;
   } POD_NameArray;
   POD_NameArray POD_Names[] =
   {
      { POD_TYPE_THERMOMETER, "Thermometer"},
      { POD_TYPE_ALARM, "Alarm"},
      { POD_TYPE_MOTORCONTROL, "Motor Controller"},
      { POD_TYPE_LIGHTSENSOR, "Light Sensor"},
      { POD_TYPE_LIGHT, "Light"},
      { POD_TYPE_POWERSENSOR, "Power Sensor"},
      { POD_TYPE_COMPASS, "Compass"},
      { POD_TYPE_DISTANCE, "Distance Sensor"},
      { POD_TYPE_BUTTONS, "Buttons"},
      { POD_TYPE_NONE, "No POD connected"},
      { POD_TYPE_UNKNOWN, "Unknown" }
   };
   for (Counter = 0 ; ; Counter ++)
   {
      if ((POD_Names[Counter].TypeNum == PODType) || (POD_Names[Counter].TypeNum == POD_TYPE_UNKNOWN))
      {
         strncpy(PODTypeString, POD_Names[Counter].String, POD_SIZE_TYPENAME);
         break;
      }
   }
   return(PODType);
}

int POD::GetURL(char *ReturnURL)
{
   if (URL != NULL)
   {
      strncpy(ReturnURL, URL, strlen(URL) + 1);
      return(POD_ERROR_NONE);
   }
   strncpy(ReturnURL, "", 1);
   return(POD_ERROR_BADDEVICE);
}

int POD::GetVersion(char *ReturnVersion)
{
   if (Version != NULL)
   {
      strncpy(ReturnVersion, Version, strlen(Version) + 1);
      return(POD_ERROR_NONE);
   }
   strncpy(ReturnVersion, "", 1);
   return(POD_ERROR_BADDEVICE);
}

int POD::GetName(char *ReturnName)
{
   int Result;
   unsigned char Data[200];
   int Response;

   if (Valid)
   {
      SendCommand(POD_COMMAND_GETNAME);
      Result = ReceiveBytes(1, Data);
      Response = Data[0];
      if (Response != POD_RESPONSE_ACK) printf("Bad response in GetName\n");
      Result = ReceiveBytes(POD_EEPROM_NAME_LEN, Data);
      strncpy(ReturnName, (char *)(Data), POD_EEPROM_NAME_LEN);
      return(POD_ERROR_NONE);
   }
   strncpy(ReturnName, "", 1);
   return(POD_ERROR_BADDEVICE);
}

int POD::SetName(char *NewName)
{
   int Result;
   unsigned char Data[200];
   int Response;
   int Counter;
   int NameLength;

   NameLength = strlen(NewName);
   if (NameLength > 15) NameLength = 15;
   SendCommand(POD_COMMAND_SETNAME);
   for (Counter = 0 ; Counter < POD_EEPROM_NAME_LEN ; Counter ++)
   {
      if (Counter >= NameLength) Result = SendByte('\0');
      else Result = SendByte(NewName[Counter]);
   }
   Result = ReceiveBytes(1, Data);
   Response = Data[0];
   if (Response != POD_RESPONSE_ACK) printf("Bad response from SetName\n");
   return(POD_ERROR_NONE);
}

POD::~POD()
{
   int Result;
#if (DEBUG_OBJECTS)
   printf("In ~POD ()\n");
#endif
   if (Valid)
   {
      Result = tcsetattr(Port, TCSANOW, &OrigTermIO);
      if (Result)
      {
         printf("Error while running tcsetattr (at close)\n\r");
      }
      close(Port);
   }
   if (DeviceFile) free(DeviceFile);
   DeviceFile = NULL;
   if (Version) free(Version);
   Version = NULL;
   if (URL) free(URL);
   URL = NULL;
}

int POD::TestMirror(int Count)
{
   int Counter;
   unsigned char Data[100];
   for (Counter = 0 ; Counter < Count ; Counter ++)
   {
      SendByte('a');
      ReceiveBytes(1, Data);
      if (Data[0] != 'a') printf("got 0x%02X ", Data[0]);
      else printf(".");
   }
   return(0);
}

int POD::TestConnection()
{
   return(Ping());
}

int POD::SetTerminalSettings()
{
   struct termios TermIO;
   int Result;
   Result = tcgetattr(Port, &OrigTermIO);
   Result = tcgetattr(Port, &TermIO);
   if (Result)
   {
      printf("Error while running tcgetattr\n\r");
   }
   cfsetispeed(&TermIO, B9600);
   cfsetospeed(&TermIO, B9600);
//   TermIO.c_cflag &= ~CSIZE;
//   TermIO.c_cflag |= (CS8 | CREAD | CLOCAL);
   TermIO.c_cflag |= (CREAD | CLOCAL);
//   TermIO.c_cflag = CS8 | CREAD | CLOCAL;
   TermIO.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG | ECHOK | ECHOCTL | ECHOKE);
//   TermIO.c_lflag &= ~(ECHOK | ECHOCTL | ECHOKE);
   TermIO.c_iflag &= ~(ICRNL | IXON);
   TermIO.c_oflag &= ~(OPOST | ONLCR);
   TermIO.c_cflag &= ~PARENB;
   TermIO.c_cflag &= ~CSTOPB;
   TermIO.c_cflag &= ~CSIZE;
   TermIO.c_cflag |= CS8;
   
   Result = tcsetattr(Port, TCSANOW, &TermIO);
   if (Result)
   {
      printf("Error while running tcsetattr\n\r");
   }
   return(POD_ERROR_NONE);
}

int POD::OpenDevice()
{
   int TempPort;
   TempPort = open(DeviceFile, O_RDWR | O_NOCTTY);
   if (TempPort == -1)
   {
//      printf("Could not open device %s\n\r", DeviceFile);
//      exit(-1);
      return (POD_ERROR_BADDEVICE);
   }
   Port = TempPort;
   Valid = 1;
   return(POD_ERROR_NONE);
}

int POD::SendBytes(int Length, unsigned char *Data)
{
   int Result;

#if (DEBUG > 2)
   printf("Preparing to write %d bytes...", Length);
   if (Length == 1) printf("0x%02X...", Data[0]);
   else printf("0x%04X...", (Data[0] << 8) | (Data[1]));
   fflush(stdout);
#endif
   Result = write(Port, Data, Length);
   if (Result != Length)
   {
      if (errno == ENODEV)
      {
         printf("Device %d:  Lost the port.  Trying again\n\r", DeviceNumber);
         /* try to reopen device */
      }
      return(-1);
   }
#if (DEBUG > 2)
   printf("done\n");
#endif
   return(Length);
}

int POD::ReceiveReady()
{
   fd_set rsel;
   int MaxFDs = 32;
   int Result;
   struct timeval tv;

   tv.tv_sec = 0;
   tv.tv_usec = 0;
   FD_ZERO(&rsel);
   FD_SET(Port, &rsel);
   Result = select(MaxFDs, &rsel, NULL, NULL, &tv);

   if (!Result) return(0);
   if (FD_ISSET(Port, &rsel) == 0) return(0);
printf("*****Something is waiting*****\n");
   return(1);
}

int POD::ReceiveBytes(int Length, unsigned char *Data)
{
   int Result;
   int Counter;
   int BytesReceived;

#if (DEBUG > 2)
   printf("Preparing to read %d bytes...", Length);
   fflush(stdout);
#endif
   BytesReceived = 0;
   for (Counter = 0 ; Counter < Length ; Counter ++)
   {
      Result = read(Port, Data + Counter, 1);
      if (Result != 1) { printf("Device %d.  error %d\n", DeviceNumber, Result); perror("here"); fflush(stdout); }
      BytesReceived ++;
#if (DEBUG > 2)
      if (isprint(Data[Counter])) printf("Received 0x%X = %c\n", Data[Counter], Data[Counter]);
      else printf("Received 0x%02X\n", Data[Counter]);
      fflush(stdout);
#endif
   }
   
   if (BytesReceived == Length) return(Length);
   if (errno == ENODEV)
   {
      printf("Device %d:  Lost the port while reading.\n\r", DeviceNumber);
      return(-1);
   }
#if (DEBUG > 2)
   printf("done\n");
#endif
   return(Length);
}

void POD::DisplayPODInfo()
{
   unsigned int TmpMyType;
   char TmpPODTypeName[POD_SIZE_TYPENAME];
   char TmpPODName[200];
   char TmpURL[200];
   char TmpVersion[200];
   int TmpDeviceNum;
                     
   TmpDeviceNum = GetDeviceNumber();
   TmpMyType = GetType(TmpPODTypeName);
   GetURL(TmpURL);
   GetName(TmpPODName);
   GetVersion(TmpVersion);

   printf("POD %d - %s named \"%s\".  URL=\"%s\", Version=\"%s\".\n",
      TmpDeviceNum, TmpPODTypeName, TmpPODName, TmpURL, TmpVersion);
}

int POD::IsValid()
{
   if (Valid == 0) return(0);
   if (PODType == POD_TYPE_NONE) return(0);
   return(1);
}

int POD::Test()
{
   printf("I'm a base POD\n");
   return(0);
}

